import numpy as np
import matplotlib.pyplot as plt


def make_bev(beam, ax):

    # Leaf対の数
    num_of_leaves = beam.BeamLimitingDeviceSequence[2].NumberOfLeafJawPairs

    cp0 = beam.ControlPointSequence[0]  # 最初のコントロールポイント

    # X Jawの位置
    jaw_x = cp0.BeamLimitingDevicePositionSequence[0].LeafJawPositions
    # Y Jawの位置
    jaw_y = cp0.BeamLimitingDevicePositionSequence[1].LeafJawPositions
    # MLCの位置
    leaf_position = cp0.BeamLimitingDevicePositionSequence[2].LeafJawPositions

    # Leaf境界のY座標
    boundaries = np.array(
        beam.BeamLimitingDeviceSequence[2].LeafPositionBoundaries)
    leaf_widths = np.diff(boundaries)  # Leafの幅

    leaf_length = 200  # Leafの長さ

    # コリメータ回転対応
    angle = float(cp0.BeamLimitingDeviceAngle)
    angle_rad = np.radians(angle)
    a_sin = np.sin(angle_rad)
    a_cos = np.cos(angle_rad)

    # X/Y Jaw
    x_0 = jaw_x[0]*a_cos - jaw_y[0]*a_sin
    y_0 = jaw_x[0]*a_sin + jaw_y[0]*a_cos
    ax.add_patch(plt.Rectangle(
        xy=(x_0, y_0),
        width=jaw_x[1]-jaw_x[0],
        height=jaw_y[1]-jaw_y[0],
        fill=False, ec='yellow', angle=angle, lw=1.5))

    for i in range(num_of_leaves):
        # X1側のMLC
        x_1 = (leaf_position[i]-leaf_length)*a_cos - boundaries[i]*a_sin
        y_1 = (leaf_position[i]-leaf_length)*a_sin + boundaries[i]*a_cos
        ax.add_patch(plt.Rectangle(
            xy=(x_1, y_1),
            width=leaf_length,
            height=leaf_widths[i],
            fill=True, ec='blue', angle=angle, lw=1., alpha=0.3))

        # X2側のMLC
        x_2 = leaf_position[i+num_of_leaves]*a_cos - boundaries[i]*a_sin
        y_2 = leaf_position[i+num_of_leaves]*a_sin + boundaries[i]*a_cos
        ax.add_patch(plt.Rectangle(
            xy=(x_2, y_2),
            width=leaf_length,
            height=leaf_widths[i],
            fill=True, ec='blue', angle=angle, lw=1., alpha=0.3))

    ax.patch.set_facecolor('gray')

    # 表示範囲の設定、最も長い対角線+marginが表示範囲に含まれるように
    disp = max([np.sqrt(i**2 + j**2) for i in jaw_x for j in jaw_y])
    margin = 30
    disp += margin
    ax.set_xlim([-disp, disp])
    ax.set_ylim([-disp, disp])
    ax.set_aspect('equal')



from .ct import CTImage
from .drr import DRR
from .bev import make_bev
from .structure import RTSS
from .dose import Dose

name = 'pydicom_tools'

__all__ = ['CTImage', 'DRR', 'make_bev', 'RTSS', 'Dose']



import os
from .. import CTImage

BASE = os.path.abspath(os.path.dirname(__file__))
SAMPLE_DIR = os.path.join(BASE, '..', 'sample')


class TestCT:
    """
    CTモジュールのテスト
    """

    def test_init(self):
        ct = CTImage()
        assert ct.is_loaded is False

    def test_load(self):
        ct = CTImage()
        ct.load(SAMPLE_DIR)

        assert ct.is_loaded is True
        assert ct.volume.shape == (5, 512, 512)



import os
import numpy as np

import pydicom


class CTImage:
    """
    CTを扱うためのクラス

    ct = CTImage()
    ct.load(path)

    でpathに存在するCTシリーズを読み込み
    """
    def __init__(self):
        """コンストラクタ"""
        self.is_loaded = False

    def load(self, path):
        """pathを指定して、CTの情報を取得"""
        if self._has_one_series(path):
            _dcm_files = self._select_cts(path)
            _ref_ct = None
            self.number_of_slices = len(_dcm_files)
            i = 0
            for dcm_file, position in sorted(_dcm_files.items(),
                                             key=lambda x: -x[1]):
                if _ref_ct is None:
                    _ref_ct = pydicom.dcmread(os.path.join(path, dcm_file))
                    self.rows = int(_ref_ct.Rows)
                    self.columns = int(_ref_ct.Columns)
                    self.volume = np.zeros(
                        (self.number_of_slices, self.rows, self.columns))
                    self.position = _ref_ct.ImagePositionPatient
                    self.pixel_spacing = _ref_ct.PixelSpacing
                    self.for_uid = _ref_ct.FrameOfReferenceUID
                    self.thickness = float(_ref_ct.SliceThickness)
                dcm_data = pydicom.dcmread(os.path.join(path, dcm_file))
                hu_value = dcm_data.pixel_array * dcm_data.RescaleSlope + \
                    dcm_data.RescaleIntercept
                self.volume[i, :, :] = hu_value
                i += 1
            self.x_min = float(self.position[0])
            self.x_max = self.x_min + \
                (self.columns - 1) * float(self.pixel_spacing[1])
            self.x_array = np.linspace(self.x_min,
                                       self.x_max,
                                       num=self.columns)
            self.y_min = float(self.position[1])
            self.y_max = self.y_min + \
                (self.rows - 1) * float(self.pixel_spacing[0])
            self.y_array = np.linspace(self.y_min, self.y_max, num=self.rows)
            self.z_min = min(_dcm_files.values())
            self.z_max = max(_dcm_files.values())
            self.z_array = np.linspace(self.z_max,
                                       self.z_min,
                                       num=self.number_of_slices)
            self.is_loaded = True
        else:
            return "There are no CT seriese or more than 2 CT seriese."

    def _select_cts(self, path):
        """指定されたディレクトリから、DICOM CTとそのスライス位置を抽出"""
        if self._has_one_series(path):
            _dcm_files = {}
            for file_path in os.listdir(path):
                if os.path.isdir(os.path.join(path, file_path)):
                    continue
                if pydicom.misc.is_dicom(os.path.join(path, file_path)):
                    dcm_data = pydicom.dcmread(os.path.join(path, file_path))
                    if dcm_data.Modality == 'CT':
                        slice_position = float(dcm_data.SliceLocation)
                        _dcm_files[file_path] = slice_position
            return _dcm_files
        else:
            return "There are no CT seriese or more than 2 CT seriese."

    def _has_one_series(self, path):
        """指定されたディレクトリにCTシリーズが1種類だけしかないか判定"""
        uids = []
        for file_path in os.listdir(path):
            if os.path.isdir(os.path.join(path, file_path)):
                continue
            if pydicom.misc.is_dicom(os.path.join(path, file_path)):
                dcm_file = pydicom.dcmread(os.path.join(path, file_path))
                if dcm_file.Modality == 'CT':
                    uids.append(dcm_file.SeriesInstanceUID)
        if len(set(uids)) == 1:
            return True
        else:
            return False



import numpy as np
from .ct import CTImage

ITK_AVAILABLE = True
try:
    import itk
except ImportError:
    ITK_AVAILABLE = False


class DRR:
    """
    DRRに関するクラス
    """
    def __init__(self, ct, beam, size=256):
        if ITK_AVAILABLE:
            self.size = size
            self.image = self._make_drr(ct, beam, size)
            self.extent = (-size/2., size/2., -size/2., size/2.)
        else:
            print(
                'You need ITK to make DRR.\n'
                'Install ITK with "pip install itk".'
            )

    def _make_drr(self, ct, beam, size):
        """
        CTImageクラスのインスタンスとBeamからDRR生成

        Parameters
        ----------
        ct : Instance of CTImage class
            CTImageクラスのインスタンス
        beam : object
            BeamSequenceの要素
        size : float
            DRR Imageのサイズ(mm)
        """
        if not isinstance(ct, CTImage):
            raise TypeError(
                "1st argument must be the instance of CTImage class.")

        # isocenter座標、ガントリ角、カウチ角度の取り出し
        cp_0 = beam.ControlPointSequence[0]  # 最初のコントロールポイント
        isocenter = cp_0.IsocenterPosition
        gan_angle = np.radians(float(cp_0.GantryAngle))
        couch_angle = np.radians(float(cp_0.PatientSupportAngle))

        # numpy配列からITK Imageに変換
        image = itk.GetImageFromArray(ct.volume.astype(np.int16))

        # 原点位置をIsocenter位置に設定
        origin = (ct.position[0] - isocenter[0],
                  ct.position[1] - isocenter[1],
                  -ct.position[2]+isocenter[2])
        # ボクセル間隔を設定
        spacing = (ct.pixel_spacing[0], ct.pixel_spacing[1], ct.thickness)

        image.SetOrigin(origin)
        image.SetSpacing(spacing)

        # RayCastの設定
        image_type = itk.Image[itk.SS, 3]
        ray_caster_type = \
            itk.RayCastInterpolateImageFunction[image_type, itk.D]
        interp = ray_caster_type.New()
        interp.SetInputImage(image)

        # 仮想線源の位置を設定 (SAD=1000 mm)
        g_sin = np.sin(gan_angle)
        g_cos = np.cos(gan_angle)
        c_sin = np.sin(couch_angle)
        c_cos = np.cos(couch_angle)

        focus = (1000*g_sin*c_cos, -1000*g_cos, 1000*g_sin*c_sin)
        interp.SetFocalPoint(focus)

        # 仮想線源位置が変更された際の変換
        transform_type = itk.TranslationTransform[itk.D, 3]
        transform = transform_type.New()
        interp.SetTransform(transform)

        # ピクセル間を線形に補間
        interpolate_type = \
            itk.LinearInterpolateImageFunction[image_type, itk.D]
        aux_interpolator = interpolate_type.New()
        interp.SetInterpolator(aux_interpolator)

        # HU値-200以上のみを反映
        interp.SetThreshold(-200)

        # 256x256ピクセルのDRRを生成
        drr = np.zeros((256, 256))

        # 1ピクセルサイズ
        pixel = size / 256.

        for i in range(256):
            for j in range(256):
                query = ((i-128)*pixel*g_cos*c_cos-(j-128)*pixel*c_sin,
                         (i-128)*pixel*g_sin,
                         (i-128)*pixel*g_cos*c_sin+(j-128)*pixel*c_cos)
                drr[j, i] = interp.Evaluate(query)

        return drr



import random
import numpy as np
from matplotlib.path import Path
import pydicom


class RTSS:
    """
    RT Structure Setを扱うためのクラス

    rtss = RTSS(path_to_rtss)

    でRTSS読み込み
    """
    def __init__(self, path):
        self._rtss = pydicom.dcmread(path)

        # Modalityチェック
        if self._rtss.Modality != 'RTSTRUCT':
            raise TypeError('This file is not DICOM RT StructureSet.')

        self.structures = {}
        for roi in self._rtss.StructureSetROISequence:
            self.structures[roi.ROINumber] = roi.ROIName

        self.points = self._get_points()
        self.paths = self._get_paths()
        self.colors = self._get_colors()

    def _get_points(self):
        _points_dict = {}

        for contour in self._rtss.ROIContourSequence:
            _structure = self.structures[contour.ReferencedROINumber]
            _points = {}
            # Emptyか確認
            if not hasattr(contour, 'ContourSequence'):
                _points_dict[_structure] = None
                continue
            for c in contour.ContourSequence:
                if c.ContourGeometricType != 'CLOSED_PLANAR':
                    continue
                _contour_data = c.ContourData
                _x = [float(x) for x in _contour_data[::3]]
                _y = [float(y) for y in _contour_data[1::3]]
                _z = float(_contour_data[2])
                _xy = list(zip(_x, _y))
                _xy.append(_xy[0])

                if _z not in _points:
                    _points[_z] = []
                _points[_z].append(_xy)

            _points_dict[_structure] = _points

        return _points_dict

    def _get_paths(self):
        _paths_dict = {}
        for contour in self._rtss.ROIContourSequence:
            _structure = self.structures[contour.ReferencedROINumber]
            _points = self.points[_structure]
            _paths = {}

            # Emptyか確認
            if _points is None:
                _paths_dict[_structure] = None
                continue

            for z, p in _points.items():
                if len(p) > 1:
                    for i, c in enumerate(p):
                        _codes = np.ones(len(c)) * Path.LINETO
                        _codes[0] = Path.MOVETO  # 輪郭点のはじめはMOVETO
                        _codes[-1] = Path.CLOSEPOLY  # 輪郭点の終わりはCLOSEPOLY
                        if i == 0:  # 最初の輪郭
                            _all_paths = c
                            _all_codes = _codes
                        else:  # 2つ目以降の輪郭を合体
                            _all_paths = np.concatenate((_all_paths, c))
                            _all_codes = np.concatenate((_all_codes, _codes))
                else:  # 同じz座標に輪郭が1つの場合
                    _all_paths = p[0]
                    _all_codes = np.ones(len(p[0])) * Path.LINETO
                    _all_codes[0] = Path.MOVETO
                    _all_codes[-1] = Path.CLOSEPOLY
                _paths[z] = Path(_all_paths, _all_codes)
            _paths_dict[_structure] = _paths

        return _paths_dict

    def _get_colors(self):
        _colors = {}
        for contour in self._rtss.ROIContourSequence:
            _structure = self.structures[contour.ReferencedROINumber]
            if hasattr(contour, 'ROIDisplayColor'):
                _color = [
                    float(contour.ROIDisplayColor[0]) / 255.,
                    float(contour.ROIDisplayColor[1]) / 255.,
                    float(contour.ROIDisplayColor[2]) / 255., 0.2
                ]
            else:
                _color = [
                    random.random() / 255.,
                    random.random() / 255.,
                    random.random() / 255., 0.2
                ]
            _colors[_structure] = _color

        return _colors

    def calc_volume(self, structure_name):
        """
        structure_name : str
        """
        if structure_name not in self.structures.values():
            raise ValueError(f"{structure_name} was not found.")

        _points = self.points[structure_name]

        if _points is None:
            raise ValueError(f"{structure_name} is empty.")

        _thickness = self._calc_thickness(_points)

        _s = 0

        for pts in _points.values():
            for p in pts:
                for i in range(len(p) - 1):
                    _s += (p[i][0] * p[i + 1][1] - p[i + 1][0] * p[i][1]) / 2.

        _v = _s * _thickness / 1000.

        return _v

    def _calc_thickness(self, _points):
        _z_list = [z for z in _points.keys()]
        _z_diff = np.diff(sorted(_z_list))

        return _z_diff.min()



import setuptools

with open("README_en.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pydicom_tools",
    version="0.0.9",
    author="Akihisa Wakita",
    author_email="wakita84@gmail.com",
    description="Tools for dicom RT analysis with pydicom",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/wkt84/pydicom-tools",
    packages=setuptools.find_packages(),
    install_requires=[
        "pydicom",
        "numpy",
        "matplotlib",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import os
from .. import RTSS

BASE = os.path.abspath(os.path.dirname(__file__))
SAMPLE_DIR = os.path.join(BASE, '..', 'sample')
RTSS_NAME = "RS.002445.dcm"


class TestStructure:
    """
    Structureモジュールのテスト
    """

    def test_init(self):
        rtss = RTSS(os.path.join(SAMPLE_DIR, RTSS_NAME))

        assert rtss.structures[1] == 'RT LENS'



import numpy as np
from scipy.interpolate import interp1d
from scipy.interpolate import RegularGridInterpolator as rgi

import pydicom


class Dose:
    def __init__(self, path):
        self._ds = pydicom.dcmread(path)
        # Modalityチェック
        if self._ds.Modality != 'RTDOSE':
            raise TypeError('This file is not DICOM RT Dose.')
        self._analyze_dicom(self._ds)

    def _analyze_dicom(self, ds):
        self.shape = ds.pixel_array.shape
        self.dose = ds.pixel_array * ds.DoseGridScaling
        self._origin = list(map(float, ds.ImagePositionPatient))
        self.x_array = np.linspace(
            self._origin[0], self._origin[0] +
            (int(ds.Columns) - 1) * float(ds.PixelSpacing[1]), int(ds.Columns))
        self.y_array = np.linspace(
            self._origin[1],
            self._origin[1] + (int(ds.Rows) - 1) * float(ds.PixelSpacing[0]),
            int(ds.Rows))
        self.z_array = np.array(ds.GridFrameOffsetVector) + self._origin[2]

    def get_2d_plane(self, v, axis):
        if axis == 'axial':
            f = interp1d(self.z_array,
                         self.dose,
                         axis=0,
                         bounds_error=False,
                         fill_value=0.)
            return f(v)
        elif axis == 'coronal':
            f = interp1d(self.y_array,
                         self.dose,
                         axis=1,
                         bounds_error=False,
                         fill_value=0.)
            return f(v)
        elif axis == 'sagittal':
            f = interp1d(self.x_array,
                         self.dose,
                         axis=2,
                         bounds_error=False,
                         fill_value=0.)
            return f(v)
        else:
            raise ValueError('No axis.')

    def get_point_dose(self, v):
        data_grid = [self.x_array, self.y_array, self.z_array]
        f = rgi(data_grid,
                self.dose.transpose(),
                bounds_error=False,
                fill_value=0)
        return f(v)
